import json
from django.conf import settings
from django.urls import reverse

from oscar.core.loading import get_model

from oscarapi.tests.utils import APITest


Basket = get_model('basket', 'Basket')


class OfferTest(APITest):
    fixtures = [
        'product', 'productcategory', 'productattribute', 'productclass',
        'productattributevalue', 'category', 'attributeoptiongroup', 
        'attributeoption', 'stockrecord', 'partner', 'offer'
    ]

    def test_basket_discount(self):
        "A discount should be properly applied to the basket total"
        self.response = self.post(
            'api-basket-add-product',
            url="http://testserver/api/products/1/",
            quantity=5)
        self.response.assertStatusEqual(200)

        self.response.assertValueEqual('total_excl_tax', "42.00")
        self.response.assertValueEqual('total_incl_tax', "42.00")
        self.response.assertValueEqual('total_excl_tax_excl_discounts', "50.00")
        self.response.assertValueEqual('total_incl_tax_excl_discounts', "50.00")

    def test_basket_line_discount(self):
        "A discount should be properly applied to the basket line"
        self.test_basket_discount()
        self.response = self.get(self.response['lines'])
        self.assertEqual(len(self.response.body), 1)
        line0 = self.response.body[0]
        self.assertEqual(line0['product'], "http://testserver/api/products/1/")
        self.assertEqual(line0['quantity'], 5)
        self.assertEqual(line0['price_excl_tax'], "42.00")
        self.assertEqual(line0['price_incl_tax'], "42.00")
        self.assertEqual(line0['price_excl_tax_excl_discounts'], "50.00")
        self.assertEqual(line0['price_incl_tax_excl_discounts'], "50.00")

        # check if the discount is applied when we retrieve a single line
        self.response = self.get(self.response.body[0]['url'])
        single_line0 = self.response.body
        self.assertEqual(single_line0['price_excl_tax'], "42.00")
        self.assertEqual(single_line0['price_incl_tax'], "42.00")
        self.assertEqual(single_line0['price_excl_tax_excl_discounts'], "50.00")
        self.assertEqual(single_line0['price_incl_tax_excl_discounts'], "50.00")

    def test_adjust_line_quantity_discount(self):
        "A single line response should return the price including the discount"
        self.response = self.post(
            'api-basket-add-product',
            url="http://testserver/api/products/1/",
            quantity=1)
        self.response.assertStatusEqual(200)

        # Get the basket lines, and update the quantity to 2
        self.response = self.get('api-basket')
        self.response = self.get(self.response['lines'])
        basket_line_url = self.response.data[0]['url']
        self.response = self.put(basket_line_url, quantity=2)
        self.response.assertStatusEqual(200)

        # this price is excl discount
        self.assertEqual(
            self.response['price_incl_tax_excl_discounts'], '20.00')
        # this price should be incl discount
        self.assertEqual(self.response['price_incl_tax'], '12.00')
        self.assertEqual(self.response['price_excl_tax'], '12.00')



from oscarapi.views.root import *
from oscarapi.views.basic import *
from oscarapi.views.login import *
from oscarapi.views.basket import *
from oscarapi.views.checkout import *
from oscarapi.views.product import *



from rest_framework.permissions import BasePermission, IsAuthenticated

from oscarapi.basket.operations import request_allows_access_to


class HasUser(BasePermission):
    "Only anonymous and authenticated users can access this resource."
    def has_permission(self, request, view):
        return request.user


class IsAdminUserOrRequestAllowsAccessTo(BasePermission):
    """
    Permission class that checks if a request allows access to a basket.
    """
    def has_object_permission(self, request, view, obj):
        return request_allows_access_to(request, obj) or request.user.is_staff


class IsOwner(IsAuthenticated):
    """
    Permission that checks if this object has a foreign key pointing to the
    authenticated user of this request
    """

    def has_object_permission(self, request, view, obj):
        return obj.user == request.user



from rest_framework import serializers
from django.utils.translation import ugettext as _

from oscarapi.utils import (
    OscarModelSerializer,
    overridable,
    OscarHyperlinkedModelSerializer
)
from oscar.core.loading import get_model


Product = get_model('catalogue', 'Product')
ProductClass = get_model('catalogue', 'ProductClass')
ProductCategory = get_model('catalogue', 'ProductCategory')
ProductAttribute = get_model('catalogue', 'ProductAttribute')
ProductAttributeValue = get_model('catalogue', 'ProductAttributeValue')
AttributeOption = get_model('catalogue', 'AttributeOption')
ProductImage = get_model('catalogue', 'ProductImage')
Option = get_model('catalogue', 'Option')
Partner = get_model('partner', 'Partner')


class PartnerSerializer(OscarModelSerializer):
    class Meta:
        model = Partner
        fields = '__all__'


class OptionSerializer(OscarHyperlinkedModelSerializer):
    class Meta:
        model = Option
        fields = overridable('OSCARAPI_OPTION_FIELDS', default=(
            'url', 'id', 'name', 'code', 'type'
        ))


class ProductAttributeValueSerializer(OscarModelSerializer):
    name = serializers.CharField(source="attribute.name")
    code = serializers.CharField(source="attribute.code")
    value = serializers.SerializerMethodField()

    def get_value(self, obj):
        obj_type = obj.attribute.type
        if obj_type == ProductAttribute.OPTION:
            return obj.value.option
        elif obj_type == ProductAttribute.MULTI_OPTION:
            return obj.value.values_list('option', flat=True)
        elif obj_type == ProductAttribute.FILE:
            return obj.value.url
        elif obj_type == ProductAttribute.IMAGE:
            return obj.value.url
        elif obj_type == ProductAttribute.ENTITY:
            if hasattr(obj.value, 'json'):
                return obj.value.json()
            else:
                return _(
                    "%(entity)s has no json method, can not convert to json" % {
                        'entity': repr(obj.value)
                    }
                )

        # return the value as stored on ProductAttributeValue in the correct type
        return obj.value

    class Meta:
        model = ProductAttributeValue
        fields = overridable(
            'OSCARAPI_PRODUCT_ATTRIBUTE_VALUE_FIELDS',
            default=('name', 'value', 'code'))


class ProductAttributeSerializer(OscarModelSerializer):
    productattributevalue_set = ProductAttributeValueSerializer(many=True)

    class Meta:
        model = ProductAttribute
        fields = overridable(
            'OSCARAPI_PRODUCT_ATTRIBUTE_FIELDS',
            default=('name', 'productattributevalue_set'))


class ProductImageSerializer(OscarModelSerializer):
    class Meta:
        model = ProductImage
        fields = '__all__'


class AvailabilitySerializer(serializers.Serializer):
    is_available_to_buy = serializers.BooleanField()
    num_available = serializers.IntegerField(required=False)
    message = serializers.CharField()


class RecommmendedProductSerializer(OscarModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name='product-detail')

    class Meta:
        model = Product
        fields = overridable(
            'OSCARAPI_RECOMMENDED_PRODUCT_FIELDS', default=('url',))


class BaseProductSerializer(OscarModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name='product-detail')
    stockrecords = serializers.HyperlinkedIdentityField(
        view_name='product-stockrecord-list')
    attributes = ProductAttributeValueSerializer(
        many=True, required=False, source="attribute_values")
    categories = serializers.StringRelatedField(many=True, required=False)
    product_class = serializers.StringRelatedField(required=False)
    price = serializers.HyperlinkedIdentityField(view_name='product-price')
    availability = serializers.HyperlinkedIdentityField(
        view_name='product-availability')
    options = OptionSerializer(many=True, required=False)
    recommended_products = RecommmendedProductSerializer(
        many=True, required=False)

    def get_field_names(self, declared_fields, info):
        """
        Override get_field_names to make sure that we are not getting errors
        for not including declared fields.
        """
        return super(BaseProductSerializer, self).get_field_names({}, info)

    class Meta:
        model = Product


class ChildProductserializer(BaseProductSerializer):
    parent = serializers.HyperlinkedRelatedField(
        view_name='product-detail', queryset=Product.objects)
    # the below fields can be filled from the parent product if enabled.
    images = ProductImageSerializer(many=True, required=False, source='parent.images')
    description = serializers.CharField(source='parent.description')

    class Meta(BaseProductSerializer.Meta):
        fields = overridable(
            'OSCARAPI_CHILDPRODUCTDETAIL_FIELDS',
            default=(
                'url', 'upc', 'id', 'title', 'structure',
                # 'parent', 'description', 'images', are not included by default, but
                # easily enabled by overriding OSCARAPI_CHILDPRODUCTDETAIL_FIELDS
                # in your settings file
                'date_created', 'date_updated', 'recommended_products',
                'attributes', 'categories', 'product_class',
                'stockrecords', 'price', 'availability', 'options'))


class ProductSerializer(BaseProductSerializer):
    images = ProductImageSerializer(many=True, required=False)
    children = ChildProductserializer(many=True, required=False)

    class Meta(BaseProductSerializer.Meta):
        fields = overridable(
            'OSCARAPI_PRODUCTDETAIL_FIELDS',
            default=(
                'url', 'upc', 'id', 'title', 'description', 'structure',
                'date_created', 'date_updated', 'recommended_products',
                'attributes', 'categories', 'product_class',
                'stockrecords', 'images', 'price', 'availability', 'options',
                'children'))


class ProductLinkSerializer(ProductSerializer):
    class Meta(BaseProductSerializer.Meta):
        fields = overridable(
            'OSCARAPI_PRODUCT_FIELDS', default=(
                'url', 'id', 'upc', 'title'
            ))


class OptionValueSerializer(serializers.Serializer):
    option = serializers.HyperlinkedRelatedField(
        view_name='option-detail', queryset=Option.objects)
    value = serializers.CharField()


class AddProductSerializer(serializers.Serializer):
    """
    Serializes and validates an add to basket request.
    """
    quantity = serializers.IntegerField(required=True)
    url = serializers.HyperlinkedRelatedField(
        view_name='product-detail', queryset=Product.objects, required=True)
    options = OptionValueSerializer(many=True, required=False)

    class Meta:
        model = Product



import logging
import re

from django.conf import settings
from django.contrib.sessions.middleware import SessionMiddleware
from django.core.exceptions import PermissionDenied
from django.urls import reverse
from django.http.response import HttpResponse
from django.utils.deprecation import MiddlewareMixin
from django.utils.translation import ugettext as _

from oscar.core.loading import get_class

from rest_framework import HTTP_HEADER_ENCODING
from rest_framework import exceptions
from rest_framework import authentication

from oscarapi.basket.operations import (
    request_allows_access_to_basket,
    store_basket_in_session,
    get_basket
)
from oscarapi.utils import (
    get_domain,
    session_id_from_parsed_session_uri,
    get_session
)
from oscarapi import models


BasketMiddleware = get_class('basket.middleware', 'BasketMiddleware')


logger = logging.getLogger(__name__)

HTTP_SESSION_ID_REGEX = re.compile(
    r'^SID:(?P<type>(?:ANON|AUTH)):(?P<realm>.*?):(?P<session_id>.+?)(?:[-:][0-9a-fA-F]+){0,2}$')


def parse_session_id(request):
    """Parse a session id from the request"""
    unparsed_session_id = request.META.get('HTTP_SESSION_ID', None)
    if unparsed_session_id is not None:
        parsed_session_id = HTTP_SESSION_ID_REGEX.match(unparsed_session_id)
        if parsed_session_id is not None:
            return parsed_session_id.groupdict()

    return None


def start_or_resume(session_id, session_type):
    if session_type == 'ANON':
        return get_session(session_id, raise_on_create=False)

    return get_session(session_id, raise_on_create=True)


class IsApiRequest(object):
    @staticmethod
    def is_api_request(request):
        path = request.path.lower()
        api_root = reverse('api-root').lower()
        return path.startswith(api_root)


class HeaderSessionMiddleware(SessionMiddleware, IsApiRequest):
    """
    Implement session through headers:

    http://www.w3.org/TR/WD-session-id

    TODO:
    Implement gateway protection, with permission options for usage of
    header sessions. With that in place the api can be used for both trusted
    and non trusted clients, see README.rst.
    """
    def process_request(self, request):
        """
        Parse the session id from the 'Session-Id: ' header when using the api.
        """
        if self.is_api_request(request):
            try:
                parsed_session_uri = parse_session_id(request)
                if parsed_session_uri is not None:
                    domain = get_domain(request)
                    if parsed_session_uri['realm'] != domain:
                        raise exceptions.NotAcceptable(
                            _('Can not accept cookie with realm %s on realm %s') % (
                                parsed_session_uri['realm'],
                                domain
                            )
                        )
                    session_id = session_id_from_parsed_session_uri(
                        parsed_session_uri)
                    request.session = start_or_resume(
                        session_id, session_type=parsed_session_uri['type'])
                    request.parsed_session_uri = parsed_session_uri

                    # since the session id is assigned by the CLIENT, there is
                    # no point in having csrf_protection. Session id's read
                    # from cookies, still need csrf!
                    request.csrf_processing_done = True
                    return None
            except exceptions.APIException as e:
                response = HttpResponse('{"reason": "%s"}' % e.detail,
                                        content_type='application/json')
                response.status_code = e.status_code
                return response

        return super(HeaderSessionMiddleware, self).process_request(request)

    def process_response(self, request, response):
        """
        Add the 'Session-Id: ' header when using the api.
        """
        if self.is_api_request(request) \
                and getattr(request, 'session', None) is not None \
                and hasattr(request, 'parsed_session_uri'):
            session_key = request.session.session_key
            parsed_session_key = session_id_from_parsed_session_uri(
                request.parsed_session_uri)
            assert(session_key == parsed_session_key), \
                '%s is not equal to %s' % (session_key, parsed_session_key)
            response['Session-Id'] = \
                'SID:%(type)s:%(realm)s:%(session_id)s' % (
                    request.parsed_session_uri)

        return super(HeaderSessionMiddleware, self).process_response(
            request, response)


class ApiGatewayMiddleWare(MiddlewareMixin, IsApiRequest):
    """
    Protect the api gateway with a token.
    """

    def process_request(self, request):
        if self.is_api_request(request):
            key = authentication.get_authorization_header(request)
            key = key.decode(HTTP_HEADER_ENCODING)
            if models.ApiKey.objects.filter(key=key).exists():
                return None

            logger.error('Invalid credentials provided for %s:%s by %s' % (
                request.method,
                request.path,
                request.META.get('REMOTE_ADDR', '<unknown>')
            ))
            raise PermissionDenied()

        return None


class ApiBasketMiddleWare(BasketMiddleware, IsApiRequest):
    """
    Use this middleware instead of Oscar's basket middleware if you
    want to mix the api with regular oscar views.

    Note that this middleware only works with MIDDLEWARE (Django > 1.10)
    as MIDDLEWARE_CLASSES is deprecated and Oscar drops the MiddlewareMixin
    compatibility in BasketMiddleware since version 1.6

    Oscar uses a cookie based session to store baskets for anonymous users, but
    oscarapi can not do that, because we don't want to put the burden
    of managing a cookie jar on oscarapi clients that are not websites.
    """
    def __call__(self, request):
        if self.is_api_request(request):
            request.cookies_to_delete = []
            # we should make sure that any cookie baskets are turned into
            # session baskets, since oscarapi uses only baskets from the
            # session.
            cookie_key = self.get_cookie_key(request)

            basket = self.get_cookie_basket(
                cookie_key,
                request,
                Exception("get_cookie_basket doesn't use the manager argument")
            )

            if basket is not None:
                # when a basket exists and we are already allowed to access
                # this basket
                if request_allows_access_to_basket(request, basket):
                    pass
                else:
                    store_basket_in_session(basket, request.session)

        return super(ApiBasketMiddleWare, self).__call__(request)

    def process_response(self, request, response):
        if self.is_api_request(request) and hasattr(request, 'user') and request.session:
            # at this point we are sure a basket can be found in the session
            # (if the session hasn't been destroyed by logging out),
            # because it is enforced in process_request.
            # We just have to make sure it is stored as a cookie, because it
            # could have been created by oscarapi.
            cookie_key = self.get_cookie_key(request)
            basket = get_basket(request)
            cookie = self.get_basket_hash(basket.id)

            # Delete any surplus cookies
            cookies_to_delete = getattr(request, 'cookies_to_delete', [])
            for cookie_key in cookies_to_delete:
                response.delete_cookie(cookie_key)

            if not request.user.is_authenticated:
                response.set_cookie(
                    cookie_key, cookie,
                    max_age=settings.OSCAR_BASKET_COOKIE_LIFETIME,
                    secure=settings.OSCAR_BASKET_COOKIE_SECURE, httponly=True)
            return response
        else:
            return super(
                ApiBasketMiddleWare, self).process_response(request, response)



from .sandbox import *  # noqa

MIGRATION_MODULES = {
    'address': None,
    'analytics': None,
    'basket': None,
    'reviews': None,
    'catalogue': None,
    'checkout': None,
    'customer': None,
    'dashboard': None,
    'offer': None,
    'order': None,
    'partner': None,
    'payment': None,
    'promotions': None,
    'search': None,
    'shipping': None,
    'voucher': None,
    'wishlists': None,
}



from django.contrib.auth import get_user_model

from oscarapi.tests.utils import APITest


User = get_user_model()


class UserTest(APITest):
    """
    Check the users are properly created bys setUp.

    Let's do it this way instead of fixtures, because it is pretty likely
    that the User model has been changed by other people, which means the
    fixtures might not fit.
    """

    def test_admin_user(self):
        "Admin user can authenticate with django via the html frontend"
        self.login('admin', 'admin')
        self.assertEqual(int(self.client.session['_auth_user_id']), 1)

    def test_non_admin_user(self):
        "Regular user can authenticate with django via the html frontend"
        self.login('nobody', 'nobody')
        self.assertEqual(int(self.client.session['_auth_user_id']), 2)

    def test_admin_header(self):
        self.hlogin('admin', 'admin', 'admin')

    def test_non_admin_header(self):
        self.hlogin('nobody', 'nobody', 'nobody')



import mock
from six import string_types

from django.urls import reverse

from oscarapi.tests.utils import APITest
from oscarapi.serializers.product import ProductLinkSerializer


class ProductListDetailSerializer(ProductLinkSerializer):
    "subclass of ProductLinkSerializer to demonstrate showing details in listview"
    class Meta(ProductLinkSerializer.Meta):
        fields = (
            'url', 'id', 'title', 'structure', 'description', 'date_created',
            'date_updated', 'recommended_products', 'attributes',
            'categories', 'product_class', 'stockrecords', 'images',
            'price', 'availability', 'options', 'children'
        )


class ProductTest(APITest):
    fixtures = [
        'product', 'productcategory', 'productattribute', 'productclass',
        'productattributevalue', 'category', 'attributeoptiongroup',
        'attributeoption', 'stockrecord', 'partner'
    ]

    def test_product_list(self):
        "Check if we get a list of products with the default attributes"
        self.response = self.get('product-list')
        self.response.assertStatusEqual(200)
        # we should have four products
        self.assertEqual(len(self.response.body), 4)
        # default we have 3 fields
        product = self.response.body[0]
        default_fields = ['id', 'url']
        for field in default_fields:
            self.assertIn(field, product)

    def test_product_list_filter(self):
        standalone_products_url = "%s?structure=standalone" % reverse('product-list')
        self.response = self.get(standalone_products_url)
        self.response.assertStatusEqual(200)
        self.assertEqual(len(self.response.body), 2)

        parent_products_url = "%s?structure=parent" % reverse('product-list')
        self.response = self.get(parent_products_url)
        self.response.assertStatusEqual(200)
        self.assertEqual(len(self.response.body), 1)

        child_products_url = "%s?structure=child" % reverse('product-list')
        self.response = self.get(child_products_url)
        self.response.assertStatusEqual(200)
        self.assertEqual(len(self.response.body), 1)

        koe_products_url = "%s?structure=koe" % reverse('product-list')
        self.response = self.get(koe_products_url)
        self.response.assertStatusEqual(200)
        self.assertEqual(len(self.response.body), 0)

    @mock.patch('oscarapi.views.product.ProductList.get_serializer_class')
    def test_productlist_detail(self, get_serializer_class):
        "The product list should be able to render the same information as the detail page"
        # setup mocks
        get_serializer_class.return_value = ProductListDetailSerializer

        # define fields to check
        product_detail_fields = (
            'url', 'id', 'title', 'structure', 'description', 'date_created',
            'date_updated', 'recommended_products', 'attributes',
            'categories', 'product_class', 'stockrecords', 'images',
            'price', 'availability', 'options', 'children'
        )

        # fetch data
        parent_products_url = "%s?structure=parent" % reverse('product-list')
        self.response = self.get(parent_products_url)

        # make assertions
        get_serializer_class.assert_called_once_with()
        self.response.assertStatusEqual(200)
        self.assertEqual(len(self.response.body), 1)

        # load product data
        products = self.response.json()
        product_with_children = products[0]

        self.assertEqual(product_with_children['structure'], 'parent',
            "since we filtered on structure=parent the list should contain items with structure=parent"
        )
        # verify all the fields are rendered in the list view
        for field in product_detail_fields:
            self.assertIn(field, product_with_children)

        children = product_with_children['children']
        self.assertEqual(len(children), 1, "There should be 1 child")
        child = children[0]
        
        self.assertEqual(child['structure'], 'child',
            "the child should have structure=child"
        )
        self.assertNotEqual(product_with_children['id'], child['id'])
        self.assertNotEqual(product_with_children['structure'], child['structure'])
        self.assertNotIn('description', child, "child should not have a description by default")
        self.assertNotIn('images', child, "child should not have images by default")

    def test_product_detail(self):
        "Check product details"
        self.response = self.get(reverse('product-detail', args=(1,)))
        self.response.assertStatusEqual(200)
        default_fields = (
            'url', 'id', 'title', 'structure', 'description', 'date_created',
            'date_updated', 'recommended_products', 'attributes',
            'categories', 'product_class', 'stockrecords', 'images',
            'price', 'availability', 'options', 'children'
        )
        for field in default_fields:
            self.assertIn(field, self.response.body)

        self.response.assertValueEqual('title', "Oscar T-shirt")

    def test_product_attribute_entity(self):
        "Entity attribute tyoe should be supported by means of json method"
        self.response = self.get(reverse('product-detail', args=(4,)))
        self.response.assertStatusEqual(200)

        self.response.assertValueEqual('title', "Entity product")

        attributes = self.response.json()['attributes']
        attributes_by_name = {a['name']:a['value'] for a in attributes}

        self.assertIsInstance(attributes_by_name['entity'], string_types)
        self.assertEqual(attributes_by_name['entity'], "<User: admin> has no json method, can not convert to json")

    def test_product_attributes(self):
        "All oscar attribute types should be supported except entity"
        self.response = self.get(reverse('product-detail', args=(3,)))
        self.response.assertStatusEqual(200)

        self.response.assertValueEqual('title', "lots of attributes")

        attributes = self.response.json()['attributes']
        attributes_by_name = {a['name']:a['value'] for a in attributes}

        # check all the attributes and their types.
        self.assertIsInstance(attributes_by_name['boolean'], bool)
        self.assertEqual(attributes_by_name['boolean'], True)
        self.assertIsInstance(attributes_by_name['date'], string_types)
        self.assertEqual(attributes_by_name['date'], "2018-01-02")
        self.assertIsInstance(attributes_by_name['datetime'], string_types)
        self.assertEqual(attributes_by_name['datetime'], "2018-01-02T10:45:00Z")
        self.assertIsInstance(attributes_by_name['file'], string_types)
        self.assertEqual(attributes_by_name['file'], '/media/images/products/2018/01/sony-xa50ES.pdf')
        self.assertIsInstance(attributes_by_name['float'], float)
        self.assertEqual(attributes_by_name['float'], 3.2)
        self.assertIsInstance(attributes_by_name['html'], string_types)
        self.assertEqual(attributes_by_name['html'], '<p>I <strong>am</strong> a test</p>')
        self.assertIsInstance(attributes_by_name['image'], string_types)
        self.assertEqual(attributes_by_name['image'], '/media/images/products/2018/01/IMG_3777.JPG')
        self.assertIsInstance(attributes_by_name['integer'], int)
        self.assertEqual(attributes_by_name['integer'], 7)
        self.assertIsInstance(attributes_by_name['multioption'], list)
        self.assertEqual(attributes_by_name['multioption'], ['Small', 'Large'])
        self.assertIsInstance(attributes_by_name['option'], string_types)
        self.assertEqual(attributes_by_name['option'], 'Small')

    def test_product_price(self):
        "See if we get the price information"
        self.response = self.get(reverse('product-price', args=(1,)))
        self.response.assertStatusEqual(200)
        self.assertIn('excl_tax', self.response.body)

        self.response = self.get(reverse('product-price', args=(999999,)))
        self.response.assertStatusEqual(404)

    def test_product_availability(self):
        "See if we get the availability information"
        self.response = self.get(reverse('product-availability', args=(1,)))
        self.response.assertStatusEqual(200)
        self.assertIn('num_available', self.response.body)

        self.response = self.get(reverse('product-availability', args=(999999,)))
        self.response.assertStatusEqual(404)



import os
from oscar import get_core_apps
from oscar import OSCAR_MAIN_TEMPLATE_DIR

ALLOWED_HOSTS = []

BASE_DIR = os.path.dirname(__file__)

OSCARAPI_BLOCK_ADMIN_API_ACCESS = False

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}

DEBUG = True

HAYSTACK_CONNECTIONS = {
    'default': {
        'ENGINE': '',
    },
}

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django.contrib.messages',
    'django.contrib.flatpages',
    'django.contrib.staticfiles',
    'rest_framework',
    'oscarapi',
    'widget_tweaks',
] + get_core_apps()

LANGUAGE_CODE = 'en-us'

LOGGING = {
    'version': 1,
    'disable_existing_loggers': True,
    'root': {
        'level': 'DEBUG' if DEBUG else 'INFO',
        'handlers': ['console', ],
    },
    'formatters': {
        'verbose': {
            'format': '%(levelname)s %(asctime)s %(module)s %(message)s',
        },
        'simple': {
            'format': '[%(asctime)s] %(message)s'
        },
    },
    'filters': {
        'require_debug_false': {
            '()': 'django.utils.log.RequireDebugFalse'
        }
    },
    'handlers': {
        'null': {
            'level': 'DEBUG',
            'class': 'logging.NullHandler',
        },
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'verbose'
        },
        'mail_admins': {
            'level': 'ERROR',
            'class': 'django.utils.log.AdminEmailHandler',
            'filters': ['require_debug_false'],
        },
    },
    'loggers': {
        # Django loggers
        'django': {
            'propagate': True,
        },
        'django.db': {
            'level': 'WARNING'
        },
        'django.request': {
            'handlers': ['mail_admins', ],
            'level': 'ERROR',
            'propagate': False
        },
        # Third party
        'south': {
            'level': 'INFO',
        },
        'sorl.thumbnail': {
            'level': 'ERROR',
        },
        # Suppress output of this debug toolbar panel
        'template_timings_panel': {
            'handlers': ['null'],
        }
    }
}

MEDIA_URL = '/media/'
MEDIA_ROOT = os.path.join(BASE_DIR, 'media')

MIDDLEWARE = (
    'django.middleware.common.CommonMiddleware',
    'oscarapi.middleware.HeaderSessionMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'oscar.apps.basket.middleware.BasketMiddleware',
    'django.contrib.flatpages.middleware.FlatpageFallbackMiddleware',
)

REST_FRAMEWORK = {
    'CHARSET': 'utf-8',
    'DEFAULT_RENDERER_CLASSES': (
        'rest_framework.renderers.JSONRenderer',
        'rest_framework.renderers.BrowsableAPIRenderer',
    )
}

ROOT_URLCONF = 'urls'

SECRET_KEY = 'li0$-gnv)76g$yf7p@(cg-^_q7j6df5cx$o-gsef5hd68phj!4'

SITE_ID = 1

STATIC_URL = '/static/'
STATIC_ROOT = os.path.join(BASE_DIR, 'static')

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [
            os.path.join(BASE_DIR, 'templates'),
            OSCAR_MAIN_TEMPLATE_DIR
        ],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.template.context_processors.i18n',
                'django.contrib.messages.context_processors.messages',
                'oscar.apps.search.context_processors.search_form',
                'oscar.apps.promotions.context_processors.promotions',
                'oscar.apps.checkout.context_processors.checkout',
                'oscar.apps.customer.notifications.context_processors.notifications',
                'oscar.core.context_processors.metadata',
            ],
        },
    },
]

TIME_ZONE = 'UTC'

USE_I18N = True
USE_L10N = True
USE_TZ = True

from oscar.defaults import *  # noqa



from copy import deepcopy

from django.contrib.auth import get_user_model
from django.urls import reverse
from django.db.transaction import atomic

from oscarapi.tests.utils import APITest

User = get_user_model()

ADDRESS = {
    "title": "Mr",
    "first_name": "Jos",
    "last_name": "Henken",
    "line1": "Roemerlaan 44",
    "line2": "",
    "line3": "",
    "line4": "Kroekingen",
    "state": "",
    "postcode": "7777KK",
    "phone_number": "+31 26 370 4887",
    "notes": "Please deliver after 5pm",
    "is_default_for_shipping": True,
    "is_default_for_billing": False,
    "country": "http://127.0.0.1:8000/api/countries/NL/",
}


class UserAddressTest(APITest):
    """
    let's check the user addresses api
    """
    fixtures = ['country']

    def test_useraddress_anonymous(self):
        url = reverse('useraddress-list')
        self.response = self.client.get(url, content_type='application/json')

        # not logged in, no glory
        self.assertEqual(self.response.status_code, 403)

    def test_useraddress_list_and_add(self):
        "Regular users can list and add own addresses"
        self.login('nobody', 'nobody')

        url = reverse('useraddress-list')
        self.response = self.client.get(url, content_type='application/json')

        # no addresses yet
        self.assertEqual(self.response.status_code, 200)
        self.assertEqual(len(self.response.data), 0)

        # let's create one
        self.response = self.post(url, **ADDRESS)
        self.assertEqual(self.response.status_code, 201)

        # And now there should be one
        self.response = self.client.get(url, content_type='application/json')
        self.assertEqual(self.response.status_code, 200)
        self.assertEqual(len(self.response.data), 1)

        # after logout we don't see anything
        self.client.logout()
        self.response = self.client.get(url, content_type='application/json')
        self.assertEqual(self.response.status_code, 403)

    # atomic is needed because of the teardown of this testcase
    @atomic
    def test_useraddress_duplicate(self):
        "Duplicate addresses raise an integrityerror, let's catch this"
        self.login('nobody', 'nobody')

        url = reverse('useraddress-list')
        self.response = self.client.get(url, content_type='application/json')

        # let's create one
        self.response = self.post(url, **ADDRESS)
        self.assertEqual(self.response.status_code, 201)

        # And now there should be one
        self.response = self.client.get(url, content_type='application/json')
        self.assertEqual(self.response.status_code, 200)
        self.assertEqual(len(self.response.data), 1)

        # now we create the same address and this is not possible
        self.response = self.post(url, **ADDRESS)
        self.assertEqual(self.response.status_code, 406)

    def test_useraddress_update_and_delete(self):
        "Regular users can update and delete their own addresses"
        self.login('nobody', 'nobody')

        url = reverse('useraddress-list')
        # let's create one
        self.response = self.post(url, **ADDRESS)

        # there is one, let's go to the detail view
        self.response = self.client.get(url, content_type='application/json')
        self.assertEqual(self.response.status_code, 200)
        self.assertEqual(len(self.response.data), 1)

        url = self.response.data[0]['url']
        self.response = self.client.get(url, content_type='application/json')
        self.assertEqual(self.response.status_code, 200)

        # test some of the fields
        self.assertEqual(self.response.data['last_name'], "Henken")
        self.assertEqual(self.response.data['is_default_for_shipping'], True)
        self.assertEqual(
            self.response.data['notes'], "Please deliver after 5pm")

        # now update the last name
        updated_address = deepcopy(ADDRESS)
        updated_address['last_name'] = 'Van Henken'
        self.response = self.put(url, **updated_address)
        self.assertEqual(self.response.status_code, 200)
        self.assertEqual(self.response.data['last_name'], "Van Henken")

        # refetching it should not make any difference
        self.response = self.client.get(url, content_type='application/json')
        self.assertEqual(self.response.data['last_name'], "Van Henken")

        # let's delete it
        self.response = self.client.delete(
            url, content_type='application/json')

        # it has been deleted, so a 204 should be the status code
        self.assertEqual(self.response.status_code, 204)

        # and it's gone
        url = reverse('useraddress-list')
        self.response = self.client.get(url, content_type='application/json')
        self.assertEqual(self.response.status_code, 200)
        self.assertEqual(len(self.response.data), 0)

    def test_address_security(self):
        "Make sure we can't access adresses of other users"
        # first login as the first user
        self.login('nobody', 'nobody')

        nobodys_address = deepcopy(ADDRESS)
        nobodys_address['last_name'] = 'Nobody'
        url = reverse('useraddress-list')
        self.response = self.post(url, **nobodys_address)

        # save the address url
        nobodys_address_url = self.response.data['url']

        # and logout
        self.client.logout()

        # and now as the other one
        self.login('somebody', 'somebody')
        url = reverse('useraddress-list')
        somebodys_address = deepcopy(ADDRESS)
        somebodys_address['last_name'] = 'Somebody'
        self.response = self.post(url, **somebodys_address)

        # save the address url
        somebodys_address_url = self.response.data['url']

        # 'somebody' should not be able to access the address
        # created by 'nobody'
        self.response = self.client.get(
            nobodys_address_url, content_type='application/json')
        self.assertEqual(self.response.status_code, 404)

        # and we can't delete it either
        self.response = self.client.delete(
            url, content_type='application/json')
        self.assertEqual(self.response.status_code, 405)

        # logout and login again with nobody
        self.client.logout()
        self.login('nobody', 'nobody')

        # 'nobody' should not be able to access the address
        # created by 'somebody'
        self.response = self.client.get(
            somebodys_address_url, content_type='application/json')
        self.assertEqual(self.response.status_code, 404)

        # and we can't update it either
        somebodys_address['last_name'] = 'Hacked by nobody'
        self.response = self.put(somebodys_address_url, **somebodys_address)
        self.assertEqual(self.response.status_code, 404)

        # now try to change the owner of my own address to 'somebody'
        self.response = self.client.get(nobodys_address_url)
        nobody_address = self.response.data
        nobody_address['last_name'] = 'Hacked by somebody'
        nobody_address['user'] = User.objects.get(username='somebody').id
        self.response = self.put(nobodys_address_url, **nobody_address)

        # oscarapi won't choke because it will just ignore the user param
        self.assertEqual(self.response.status_code, 200)
        self.response = self.client.get(nobodys_address_url)

        # so in the end we only updated our own address
        self.assertEqual(self.response.data['last_name'], 'Hacked by somebody')

        # to be sure: somebody has only one address
        self.client.logout()
        self.login('somebody', 'somebody')

        url = reverse('useraddress-list')
        self.response = self.client.get(url, content_type='application/json')
        self.assertEqual(self.response.status_code, 200)
        self.assertEqual(len(self.response.data), 1)



from django.core.exceptions import PermissionDenied
from django.urls import reverse
from django.test import RequestFactory, TestCase

from oscarapi.middleware import ApiGatewayMiddleWare, parse_session_id
from oscarapi.models import ApiKey


class DummyRequest:
    META = {'HTTP_SESSION_ID': None}


class ApiGatewayMiddleWareTest(TestCase):

    rf = RequestFactory()

    def setUp(self):
        super(ApiGatewayMiddleWareTest, self).setUp()

        ApiKey.objects.create(key='testapikey')

    def tearDown(self):
        ApiKey.objects.filter(key='testapikey').delete()

        super(ApiGatewayMiddleWareTest, self).tearDown()

    def test_process_request(self):
        basket_url = reverse('api-basket')

        # without Authorization header
        request = self.rf.get(basket_url)
        with self.assertRaises(PermissionDenied):
            ApiGatewayMiddleWare().process_request(request)

        # invalid Authorization header
        request = self.rf.get(basket_url, HTTP_AUTHORIZATION='wrongkey')
        with self.assertRaises(PermissionDenied):
            ApiGatewayMiddleWare().process_request(request)

        # valid Authorization header
        request = self.rf.get(basket_url, HTTP_AUTHORIZATION='testapikey')
        self.assertIsNone(ApiGatewayMiddleWare().process_request(request))

    def test_parse_session_id(self):
        dummy_request = DummyRequest()

        dummy_request.META['HTTP_SESSION_ID'] = 'SID:ANON:example.com:987171879'
        self.assertEqual(
            sorted(parse_session_id(dummy_request).items()),
            [('realm', 'example.com'), ('session_id', '987171879'), ('type', 'ANON')])

        dummy_request.META['HTTP_SESSION_ID'] = 'SID:AUTH:example.com:987171879'
        self.assertEqual(
            sorted(parse_session_id(dummy_request).items()),
            [('realm', 'example.com'), ('session_id', '987171879'), ('type', 'AUTH')])

        dummy_request.META['HTTP_SESSION_ID'] = 'SID:ANON:example.com:987171879-16EF'
        self.assertEqual(
            sorted(parse_session_id(dummy_request).items()),
            [('realm', 'example.com'), ('session_id', '987171879'), ('type', 'ANON')])

        dummy_request.META['HTTP_SESSION_ID'] = 'SID:ANON:example.com:98717-16EF:100'
        self.assertEqual(
            sorted(parse_session_id(dummy_request).items()),
            [('realm', 'example.com'), ('session_id', '98717'), ('type', 'ANON')])

        dummy_request.META['HTTP_SESSION_ID'] = 'SID:ANON::987171879'
        self.assertEqual(
            sorted(parse_session_id(dummy_request).items()),
            [('realm', ''), ('session_id', '987171879'), ('type', 'ANON')])

        dummy_request.META['HTTP_SESSION_ID'] = 'SID:ANON:example.com:923-thread1'
        self.assertEqual(
            sorted(parse_session_id(dummy_request).items()),
            [('realm', 'example.com'), ('session_id', '923-thread1'), ('type', 'ANON')])

        dummy_request.META['HTTP_SESSION_ID'] = 'SID:BULLSHIT:example.com:987171879'
        self.assertIsNone(parse_session_id(dummy_request))

        dummy_request.META['HTTP_SESSION_ID'] = 'ENTIREGABRBAGE'
        self.assertIsNone(parse_session_id(dummy_request))

        dummy_request.META['HTTP_SESSION_ID'] = 'SID:ANON:987171879'
        self.assertIsNone(parse_session_id(dummy_request))



from setuptools import find_packages, setup

__version__ = "1.5.1"


setup(
    # package name in pypi
    name='django-oscar-api',
    # extract version from module.
    version=__version__,
    description="REST API module for django-oscar",
    long_description=open('README.rst').read(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
    ],
    keywords='',
    author='Lars van de Kerkhof, Martijn Jacobs',
    author_email='lars@permanentmarkers.nl, martijn@devopsconsulting.nl',
    url='https://github.com/django-oscar/django-oscar-api',
    license='BSD',
    # include all packages in the egg, except the test package.
    packages=find_packages(
        exclude=['ez_setup', 'examples', '*tests', '*fixtures', 'sandbox']),
    # for avoiding conflict have one namespace for all apc related eggs.
    namespace_packages=[],
    # include non python files
    include_package_data=True,
    zip_safe=False,
    # specify dependencies
    install_requires=[
        'setuptools',
        'django-oscar>=1.5.1',
        'djangorestframework>=3.4',
        'six'
    ],
    # mark test target to require extras.
    extras_require={
        'dev': ['coverage', 'mock', 'twine'],
        'docs': ['sphinx', 'sphinx_rtd_theme'],
    },
)



import functools

from django.contrib import auth

from oscar.core.loading import get_class, get_model

from rest_framework import generics
from rest_framework.permissions import IsAdminUser

from six.moves import map

from oscarapi import permissions, serializers
from oscarapi.basket.operations import assign_basket_strategy


from .mixin import PutIsPatchMixin

Selector = get_class('partner.strategy', 'Selector')

__all__ = (
    'BasketList', 'BasketDetail',
    'LineAttributeList', 'LineAttributeDetail',
    'StockRecordList', 'StockRecordDetail',
    'UserList', 'UserDetail',
    'OptionList', 'OptionDetail',
    'CountryList', 'CountryDetail',
    'PartnerList', 'PartnerDetail',
)

Basket = get_model('basket', 'Basket')
LineAttribute = get_model('basket', 'LineAttribute')
Product = get_model('catalogue', 'Product')
StockRecord = get_model('partner', 'StockRecord')
Option = get_model('catalogue', 'Option')
User = auth.get_user_model()
Country = get_model('address', 'Country')
Partner = get_model('partner', 'Partner')


# TODO: For all API's in this file, the permissions should be checked if they
# are sensible.
class CountryList(generics.ListAPIView):
    serializer_class = serializers.CountrySerializer
    queryset = Country.objects.all()


class CountryDetail(generics.RetrieveAPIView):
    serializer_class = serializers.CountrySerializer
    queryset = Country.objects.all()


class BasketList(generics.ListCreateAPIView):
    serializer_class = serializers.BasketSerializer
    queryset = Basket.objects.all()
    permission_classes = (IsAdminUser,)

    def get_queryset(self):
        qs = super(BasketList, self).get_queryset()
        return list(map(
            functools.partial(assign_basket_strategy, request=self.request),
            qs))


class BasketDetail(PutIsPatchMixin, generics.RetrieveUpdateDestroyAPIView):
    serializer_class = serializers.BasketSerializer
    permission_classes = (permissions.IsAdminUserOrRequestAllowsAccessTo,)
    queryset = Basket.objects.all()

    def get_object(self):
        basket = super(BasketDetail, self).get_object()
        return assign_basket_strategy(basket, self.request)


class LineAttributeList(generics.ListCreateAPIView):
    queryset = LineAttribute.objects.all()
    serializer_class = serializers.LineAttributeSerializer


class LineAttributeDetail(PutIsPatchMixin, generics.RetrieveUpdateAPIView):
    queryset = LineAttribute.objects.all()
    serializer_class = serializers.LineAttributeSerializer
    permission_classes = (permissions.IsAdminUserOrRequestAllowsAccessTo,)  # noqa


class StockRecordList(generics.ListAPIView):
    serializer_class = serializers.StockRecordSerializer
    queryset = StockRecord.objects.all()

    def get(self, request, pk=None, *args, **kwargs):
        if pk is not None:
            self.queryset = self.queryset.filter(product__id=pk)

        return super(StockRecordList, self).get(request, *args, **kwargs)


class StockRecordDetail(generics.RetrieveAPIView):
    queryset = StockRecord.objects.all()
    serializer_class = serializers.StockRecordSerializer


class UserList(generics.ListAPIView):
    queryset = User.objects.all()
    serializer_class = serializers.UserSerializer
    permission_classes = (IsAdminUser,)


class UserDetail(generics.RetrieveAPIView):
    queryset = User.objects.all()
    serializer_class = serializers.UserSerializer
    permission_classes = (IsAdminUser,)


class OptionList(generics.ListAPIView):
    queryset = Option.objects.all()
    serializer_class = serializers.OptionSerializer


class OptionDetail(generics.RetrieveAPIView):
    queryset = Option.objects.all()
    serializer_class = serializers.OptionSerializer


class PartnerList(generics.ListAPIView):
    queryset = Partner.objects.all()
    serializer_class = serializers.PartnerSerializer


class PartnerDetail(generics.RetrieveAPIView):
    queryset = Partner.objects.all()
    serializer_class = serializers.PartnerSerializer



from rest_framework import generics
from rest_framework.response import Response

from oscar.core.loading import get_class, get_model

from oscarapi import serializers


Selector = get_class('partner.strategy', 'Selector')

__all__ = (
    'ProductList', 'ProductDetail',
    'ProductPrice', 'ProductAvailability',
)

Product = get_model('catalogue', 'Product')


class ProductList(generics.ListAPIView):
    queryset = Product.objects.all()
    serializer_class = serializers.ProductLinkSerializer

    def get_queryset(self):
        """
        Allow filtering on structure so standalone and parent products can
        be selected separately, eg::

            http://127.0.0.1:8000/api/products/?structure=standalone

        or::

            http://127.0.0.1:8000/api/products/?structure=parent
        """
        qs = super(ProductList, self).get_queryset()
        structure = self.request.query_params.get('structure')
        if structure is not None:
            return qs.filter(structure=structure)

        return qs


class ProductDetail(generics.RetrieveAPIView):
    queryset = Product.objects.all()
    serializer_class = serializers.ProductSerializer


class ProductPrice(generics.RetrieveAPIView):
    queryset = Product.objects.all()

    def get(self, request, pk=None, format=None):
        product = self.get_object()
        strategy = Selector().strategy(request=request, user=request.user)
        ser = serializers.PriceSerializer(
            strategy.fetch_for_product(product).price,
            context={'request': request})
        return Response(ser.data)


class ProductAvailability(generics.RetrieveAPIView):
    queryset = Product.objects.all()

    def get(self, request, pk=None, format=None):
        product = self.get_object()
        strategy = Selector().strategy(request=request, user=request.user)
        ser = serializers.AvailabilitySerializer(
            strategy.fetch_for_product(product).availability,
            context={'request': request})
        return Response(ser.data)



from django.contrib.auth import get_user_model, authenticate
from rest_framework import serializers

from oscarapi.utils import overridable


User = get_user_model()


def field_length(fieldname):
    field = next(
        field for field in User._meta.fields if field.name == fieldname)
    return field.max_length


class UserSerializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = overridable('OSCARAPI_USER_FIELDS', (
            User.USERNAME_FIELD, 'id', 'date_joined',))


class LoginSerializer(serializers.Serializer):
    username = serializers.CharField(
        max_length=field_length(User.USERNAME_FIELD), required=True)
    password = serializers.CharField(
        max_length=field_length('password'), required=True)

    def validate(self, attrs):
        user = authenticate(
            username=attrs['username'], password=attrs['password'])
        if user is None:
            raise serializers.ValidationError('invalid login')
        elif not user.is_active:
            raise serializers.ValidationError(
                'Can not log in as inactive user')
        elif user.is_staff and overridable(
                'OSCARAPI_BLOCK_ADMIN_API_ACCESS', True):
            raise serializers.ValidationError(
                'Staff users can not log in via the rest api')

        # set instance to the user so we can use this in the view
        self.instance = user
        return attrs



# -*- coding: utf-8 -*-
import unittest
from mock import patch

from oscar.core.loading import get_model

from rest_framework.response import Response
from oscarapi.tests.utils import APITest


Basket = get_model('basket', 'Basket')


class CheckOutTest(APITest):
    fixtures = [
        'product', 'productcategory', 'productattribute', 'productclass',
        'productattributevalue', 'category', 'attributeoptiongroup',
        'attributeoption', 'stockrecord', 'partner', 'orderanditemcharges',
        'country']

    def test_checkout(self):
        """Test if an order can be placed as an authenticated user with session based auth."""
        self.login(username='nobody', password='nobody')
        response = self.get('api-basket')
        self.assertTrue(response.status_code, 200)
        basket = response.data
        basket_url = basket.get('url')
        basket_id = basket.get('id')

        request = {
            'basket': basket_url,
            'guest_email': 'henk@example.com',
            'total': '50.0',
            'shipping_method_code': "no-shipping-required",
            'shipping_charge': {
                'currency': 'EUR',
                'excl_tax': '0.00',
                'tax': '0.00'
            },
            "shipping_address": {
                "country": "http://127.0.0.1:8000/api/countries/NL/",
                "first_name": "Henk",
                "last_name": "Van den Heuvel",
                "line1": "Roemerlaan 44",
                "line2": "",
                "line3": "",
                "line4": "Kroekingen",
                "notes": "Niet STUK MAKEN OK!!!!",
                "phone_number": "+31 26 370 4887",
                "postcode": "7777KK",
                "state": "Gerendrecht",
                "title": "Mr"
            }
        }
        response = self.post('api-checkout', **request)
        self.assertEqual(response.status_code, 406)
        response = self.post(
            'api-basket-add-product',
            url="http://testserver/api/products/1/", quantity=5)
        self.assertEqual(response.status_code, 200)
        response = self.post('api-checkout', **request)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.data['guest_email'],
            '',
            'Guest email should be blank since user was authenticated')
        self.assertEqual(Basket.objects.get(pk=basket_id).status, 'Frozen', 'Basket should be frozen after placing order and before payment')

    def test_checkout_implicit_shipping(self):
        "Test if an order can be placed without specifying shipping method."
        self.login(username='nobody', password='nobody')
        response = self.get('api-basket')
        self.assertTrue(response.status_code, 200)
        basket = response.data
        basket_url = basket.get('url')
        basket_id = basket.get('id')

        request = {
            'basket': basket_url,
            'total': '50.0',
            "shipping_address": {
                "country": "http://127.0.0.1:8000/api/countries/NL/",
                "first_name": "Henk",
                "last_name": "Van den Heuvel",
                "line1": "Roemerlaan 44",
                "line2": "",
                "line3": "",
                "line4": "Kroekingen",
                "notes": "Niet STUK MAKEN OK!!!!",
                "phone_number": "+31 26 370 4887",
                "postcode": "7777KK",
                "state": "Gerendrecht",
                "title": "Mr"
            }
        }
        response = self.post('api-checkout', **request)
        self.assertEqual(response.status_code, 406)
        response = self.post('api-basket-add-product', url="http://testserver/api/products/1/", quantity=5)
        self.assertEqual(response.status_code, 200)
        response = self.post('api-checkout', **request)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(Basket.objects.get(pk=basket_id).status, 'Frozen', 'Basket should be frozen after placing order and before payment')

    def test_checkout_billing_address(self):
        "Test if an order can be placed with a billing address"
        self.login(username='nobody', password='nobody')
        response = self.get('api-basket')
        self.assertTrue(response.status_code, 200)
        basket = response.data
        basket_url = basket.get('url')

        request = {
            'basket': basket_url,
            'total': '50.0',
            "shipping_address": {
                "country": "http://127.0.0.1:8000/api/countries/NL/",
                "first_name": "Henk",
                "last_name": "Van den Heuvel",
                "line1": "Roemerlaan 44",
                "line2": "",
                "line3": "",
                "line4": "Kroekingen",
                "notes": "Niet STUK MAKEN OK!!!!",
                "phone_number": "+31 26 370 4887",
                "postcode": "7777KK",
                "state": "Gerendrecht",
                "title": "Mr"
            },
            "billing_address": {
                "country": "http://127.0.0.1:8000/api/countries/NL/",
                "first_name": "Jos",
                "last_name": "Henken",
                "line1": "Stationstraat 4",
                "line2": "",
                "line3": "",
                "line4": "Hengelo",
                "notes": "",
                "phone_number": "+31 26 370 1111",
                "postcode": "1234AA",
                "state": "Gelderland",
                "title": "Mr"
            }
        }
        self.post(
            'api-basket-add-product',
            url="http://testserver/api/products/1/", quantity=5)
        response = self.post('api-checkout', **request)
        self.assertEqual(response.status_code, 200)

    def test_checkout_wrong_billing_address(self):
        "Test if an order can be placed with a billing address"
        self.login(username='nobody', password='nobody')
        response = self.get('api-basket')
        self.assertTrue(response.status_code, 200)
        basket = response.data
        basket_url = basket.get('url')

        request = {
            'basket': basket_url,
            'total': '50.0',
            "shipping_address": {
                "country": "http://127.0.0.1:8000/api/countries/NL/",
                "first_name": "Henk",
                "last_name": "Van den Heuvel",
                "line1": "Roemerlaan 44",
                "line2": "",
                "line3": "",
                "line4": "Kroekingen",
                "notes": "Niet STUK MAKEN OK!!!!",
                "phone_number": "+31 26 370 4887",
                "postcode": "7777KK",
                "state": "Gerendrecht",
                "title": "Mr"
            },
            "billing_address": {
                "country": "This is wrong"
            }
        }
        self.post(
            'api-basket-add-product',
            url="http://testserver/api/products/1/", quantity=5)
        response = self.post('api-checkout', **request)
        # It should complain about the billing address
        self.assertEqual(response.status_code, 406)
        self.assertEqual(
            response.data['billing_address']['country'][0],
            "Invalid hyperlink - No URL match.")

    def test_checkout_total_error(self):
        "When sending a wrong total, checkout should raise an error"
        self.login(username='nobody', password='nobody')
        response = self.get('api-basket')
        self.assertTrue(response.status_code, 200)
        basket = response.data
        basket_url = basket.get('url')

        request = {
            'basket': basket_url,
            'total': '150.0',
            "shipping_address": {
                "country": "http://127.0.0.1:8000/api/countries/NL/",
                "first_name": "Henk",
                "last_name": "Van den Heuvel",
                "line1": "Roemerlaan 44",
                "line2": "",
                "line3": "",
                "line4": "Kroekingen",
                "notes": "Niet STUK MAKEN OK!!!!",
                "phone_number": "+31 26 370 4887",
                "postcode": "7777KK",
                "state": "Gerendrecht",
                "title": "Mr"
            }
        }
        self.response = self.post('api-basket-add-product', url="http://testserver/api/products/1/", quantity=5)
        self.response.assertStatusEqual(200)
        self.response = self.post('api-checkout', **request)
        self.response.assertStatusEqual(406)
        self.assertTrue("Total incorrect" in self.response['non_field_errors'][0])

    def test_utf8_encoding(self):
        "We should accept utf-8 (non ascii) characters in the address"
        self.login(username='nobody', password='nobody')
        response = self.get('api-basket')
        self.assertTrue(response.status_code, 200)
        basket = response.data
        basket_url = basket.get('url')

        request = {
            'basket': basket_url,
            "shipping_address": {
                "country": "http://127.0.0.1:8000/api/countries/NL/",
                "first_name": "Henk",
                "last_name": "Van den Heuvel",
                "line1": "Ї ❤ chǼractɇɌȘ",
                "line2": "",
                "line3": "",
                "line4": "Kroekingen",
                "notes": "Niet STUK MAKEN OK!!!!",
                "phone_number": "+31 26 370 4887",
                "postcode": "7777KK",
                "state": "Gerendrecht",
                "title": "Mr"
            }
        }
        self.response = self.post('api-basket-add-product', url="http://testserver/api/products/1/", quantity=5)
        self.response.assertStatusEqual(200)
        self.response = self.post('api-checkout', **request)
        self.response.assertStatusEqual(200)
        self.assertEqual(
            self.response.data['shipping_address']['line1'], u"Ї ❤ chǼractɇɌȘ")

    def test_checkout_empty_basket(self):
        "When basket is empty, checkout should raise an error"
        self.login(username='nobody', password='nobody')
        response = self.get('api-basket')
        self.assertTrue(response.status_code, 200)
        basket = response.data

        response = self.get(basket['lines'])
        self.assertTrue(response.status_code, 200)
        lines = response.data

        self.assertEqual(lines, [])

        request = {
            'basket': basket.get('url'),
            'total': '0.00',
            "shipping_address": {
                "country": "http://127.0.0.1:8000/api/countries/NL/",
                "first_name": "Henk",
                "last_name": "Van den Heuvel",
                "line1": "Roemerlaan 44",
                "line2": "",
                "line3": "",
                "line4": "Kroekingen",
                "notes": "Niet STUK MAKEN OK!!!!",
                "phone_number": "+31 26 370 4887",
                "postcode": "7777KK",
                "state": "Gerendrecht",
                "title": "Mr"
            }
        }
        self.response = self.post('api-checkout', **request)
        self.response.assertStatusEqual(406)
        self.response.assertValueEqual('non_field_errors', ['Cannot checkout with empty basket'])

    def test_total_is_optional(self):
        "Total should be an optional value"
        self.login(username='nobody', password='nobody')
        response = self.get('api-basket')
        self.assertTrue(response.status_code, 200)
        basket = response.data
        basket_url = basket.get('url')
        basket_id = basket.get('id')

        request = {
            'basket': basket_url,
            "shipping_address": {
                "country": "http://127.0.0.1:8000/api/countries/NL/",
                "first_name": "Henk",
                "last_name": "Van den Heuvel",
                "line1": "Roemerlaan 44",
                "line2": "",
                "line3": "",
                "line4": "Kroekingen",
                "notes": "Niet STUK MAKEN OK!!!!",
                "phone_number": "+31 26 370 4887",
                "postcode": "7777KK",
                "state": "Gerendrecht",
                "title": "Mr"
            }
        }
        self.response = self.post('api-basket-add-product', url="http://testserver/api/products/1/", quantity=5)
        self.response.assertStatusEqual(200)
        self.response = self.post('api-checkout', **request)
        self.response.assertStatusEqual(200)

    def test_can_login_with_frozen_user_basket(self):
        "When a user has an unpaid order, he should still be able to log in"
        self.test_checkout()
        self.delete('api-login')
        self.get('api-basket')
        self.post('api-basket-add-product', url="http://testserver/api/products/1/", quantity=5)
        self.response = self.post('api-login', username='nobody', password='nobody')
        self.response.assertStatusEqual(200)
        self.login(username='nobody', password='nobody')

    def test_checkout_creates_an_order(self):
        "After checkout has been done, a user should have gained an order object"
        self.test_checkout()
        self.response = self.get('order-list')
        self.assertEqual(len(self.response), 1, 'An order should have been created.')

    @unittest.skip('Please add implementation')
    def test_checkout_header(self):
        "Prove that the user 'nobody' can checkout his cart when authenticating with header session"
        self.fail('Please add implementation')

    def test_anonymous_checkout(self):
        "Test if an order can be placed as an anonymous user."
        response = self.get('api-basket')
        self.assertTrue(response.status_code, 200)
        basket = response.data
        basket_url = basket.get('url')
        basket_id = basket.get('id')

        request = {
            'basket': basket_url,
            'total': '50.0',
            'shipping_method_code': "no-shipping-required",
            'shipping_charge': {
                'currency': 'EUR',
                'excl_tax': '0.00',
                'tax': '0.00'
            },
            "shipping_address": {
                "country": "http://127.0.0.1:8000/api/countries/NL/",
                "first_name": "Henk",
                "last_name": "Van den Heuvel",
                "line1": "Roemerlaan 44",
                "line2": "",
                "line3": "",
                "line4": "Kroekingen",
                "notes": "Niet STUK MAKEN OK!!!!",
                "phone_number": "+31 26 370 4887",
                "postcode": "7777KK",
                "state": "Gerendrecht",
                "title": "Mr"
            }
        }

        with self.settings(OSCAR_ALLOW_ANON_CHECKOUT=True):
            response = self.post('api-checkout', **request)
            self.assertEqual(response.status_code, 406)
            response = self.post('api-basket-add-product', url="http://testserver/api/products/1/", quantity=5)
            self.assertEqual(response.status_code, 200)

            # no guest email specified should say 406
            response = self.post('api-checkout', **request)
            self.assertEqual(response.status_code, 406)

            # an empty email address should say this as well
            request['guest_email'] = ''
            response = self.post('api-checkout', **request)
            self.assertEqual(response.status_code, 406)

            # Add in guest_email to get a 200
            request['guest_email'] = 'henk@example.com'
            response = self.post('api-checkout', **request)
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.data['guest_email'], 'henk@example.com')
            self.assertEqual(Basket.objects.get(pk=basket_id).status, 'Frozen', 'Basket should be frozen after placing order and before payment')

    @patch('oscarapi.signals.oscarapi_post_checkout.send')
    def test_post_checkout_signal_send(self, mock):
        "The oscarapi_post_checkout signal should be send after checkout"
        self.test_anonymous_checkout()
        self.assertTrue(mock.called)
        # make sure it's a django Response instance and not the DRF module
        self.assertTrue(isinstance(mock.call_args[1]['response'], Response))

    def test_checkout_permissions(self):
        "Prove that someone can not check out someone elses cart by mistake"

        # first login as nobody
        self.login(username='nobody', password='nobody')
        response = self.get('api-basket')

        # store this basket because somebody is going to checkout with this
        basket = response.data
        nobody_basket_url = basket.get('url')

        response = self.post(
            'api-basket-add-product',
            url="http://testserver/api/products/1/", quantity=5)

        self.client.logout()

        # now login as smomebody and fill another basket
        self.login(username='somebody', password='somebody')
        response = self.post(
            'api-basket-add-product',
            url="http://testserver/api/products/1/", quantity=5)

        # so let's checkout with nobody's basket WHAHAAAHAHA!
        request = {
            'basket': nobody_basket_url,
            'total': '50.0',
            'shipping_method_code': "no-shipping-required",
            'shipping_charge': {
                'currency': 'EUR',
                'excl_tax': '0.00',
                'tax': '0.00'
            },
            "shipping_address": {
                "country": "http://127.0.0.1:8000/api/countries/NL/",
                "first_name": "Henk",
                "last_name": "Van den Heuvel",
                "line1": "Roemerlaan 44",
                "line2": "",
                "line3": "",
                "line4": "Kroekingen",
                "notes": "Niet STUK MAKEN OK!!!!",
                "phone_number": "+31 26 370 4887",
                "postcode": "7777KK",
                "state": "Gerendrecht",
                "title": "Mr"
            }
        }

        # Oh, this is indeed not possible
        response = self.post('api-checkout', **request)
        self.assertEqual(response.status_code, 401)
        self.assertEqual(response.data, "Unauthorized")

    def test_shipping_methods(self):
        """Test if shipping methods can be fetched for baskets."""
        self.login(username='nobody', password='nobody')
        response = self.get('api-basket')
        self.assertTrue(response.status_code, 200)
        basket = response.data
        basket_url = basket.get('url')
        basket_id = basket.get('id')

        request =  {
            "country": "http://127.0.0.1:8000/api/countries/NL/",
            "first_name": "Henk",
            "last_name": "Van den Heuvel",
            "line1": "Roemerlaan 44",
            "line2": "",
            "line3": "",
            "line4": "Kroekingen",
            "notes": "Niet STUK MAKEN OK!!!!",
            "phone_number": "+31 26 370 4887",
            "postcode": "7777KK",
            "state": "Gerendrecht",
            "title": "Mr"
        }
        self.response = self.post('api-basket-shipping-methods', **request)
        self.response.assertStatusEqual(200)
        self.assertEqual(len(self.response), 1)
        self.assertDictEqual(self.response[0], {
            'code': 'no-shipping-required',
            'name': 'No shipping required',
            'description': "",
            'is_discounted': False,
            'discount': 0,
            'price': {
                'currency': None,
                'excl_tax': '0.00',
                'incl_tax': '0.00',
                'tax': '0.00'
            }
        })
        response = self.post(
            'api-basket-add-product',
            url="http://testserver/api/products/1/", quantity=5)
        self.assertEqual(response.status_code, 200)

        self.response = self.post('api-basket-shipping-methods', **request)
        self.response.assertStatusEqual(200)
        self.assertEqual(len(self.response), 1)
        self.assertDictEqual(self.response[0], {
            'code': 'free-shipping',
            'name': 'Free shipping',
            'description': "",
            'is_discounted': False,
            'discount': 0,
            'price': {
                'currency': 'EUR',
                'excl_tax': '0.00',
                'incl_tax': '0.00',
                'tax': '0.00'
            }
        })


    @unittest.skip('Please add implementation')
    def test_cart_immutable_after_checkout(self):
        "Prove that the cart can not be changed with the webservice by users in any way after checkout has been made."
        self.fail("It might be that admin users can actually still modify a checked out cart (frozen)")

    @unittest.skip('Please add implementation')
    def test_client_can_not_falsify_picing(self):
        "Prove that the total and shippingcharge variable sent along with a checkout request, can not be manipulated"
        self.fail('checkout variable can not be used to buy products for prices they are not available at.')



from django.conf.urls import url

from rest_framework.urlpatterns import format_suffix_patterns

from oscarapi import views

urlpatterns = [
    url(r'^$', views.api_root, name='api-root'),
    url(r'^login/$', views.LoginView.as_view(), name='api-login'),
    url(r'^basket/$', views.BasketView.as_view(), name='api-basket'),
    url(r'^basket/add-product/$', views.AddProductView.as_view(), name='api-basket-add-product'),
    url(r'^basket/add-voucher/$', views.AddVoucherView.as_view(), name='api-basket-add-voucher'),
    url(r'^basket/shipping-methods/$', views.ShippingMethodView.as_view(), name='api-basket-shipping-methods'),
    url(r'^baskets/$', views.BasketList.as_view(), name='basket-list'),
    url(r'^baskets/(?P<pk>[0-9]+)/$', views.BasketDetail.as_view(), name='basket-detail'),
    url(r'^baskets/(?P<pk>[0-9]+)/lines/$', views.LineList.as_view(), name='basket-lines-list'),
    url(r'^baskets/(?P<basket_pk>[0-9]+)/lines/(?P<pk>[0-9]+)/$', views.BasketLineDetail.as_view(), name='basket-line-detail'),
    url(r'^lines/$', views.LineList.as_view(), name='line-list'),
    url(r'^lines/(?P<pk>[0-9]+)/$', views.LineDetail.as_view(), name='line-detail'),
    url(r'^lineattributes/$', views.LineAttributeList.as_view(), name='lineattribute-list'),
    url(r'^lineattributes/(?P<pk>[0-9]+)/$', views.LineAttributeDetail.as_view(), name='lineattribute-detail'),
    url(r'^products/$', views.ProductList.as_view(), name='product-list'),
    url(r'^products/(?P<pk>[0-9]+)/$', views.ProductDetail.as_view(), name='product-detail'),
    url(r'^products/(?P<pk>[0-9]+)/price/$', views.ProductPrice.as_view(), name='product-price'),
    url(r'^products/(?P<pk>[0-9]+)/availability/$', views.ProductAvailability.as_view(), name='product-availability'),
    url(r'^products/(?P<pk>[0-9]+)/stockrecords/$', views.StockRecordList.as_view(), name='product-stockrecord-list'),
    url(r'^stockrecords/$', views.StockRecordList.as_view(), name='stockrecord-list'),
    url(r'^stockrecords/(?P<pk>[0-9]+)/$', views.StockRecordDetail.as_view(), name='stockrecord-detail'),
    url(r'^options/$', views.OptionList.as_view(), name='option-list'),
    url(r'^options/(?P<pk>[0-9]+)/$', views.OptionDetail.as_view(), name='option-detail'),
    url(r'^users/$', views.UserList.as_view(), name='user-list'),
    url(r'^users/(?P<pk>[0-9]+)/$', views.UserDetail.as_view(), name='user-detail'),
    url(r'^checkout/$', views.CheckoutView.as_view(), name='api-checkout'),
    url(r'^orders/$', views.OrderList.as_view(), name='order-list'),
    url(r'^orders/(?P<pk>[0-9]+)/$', views.OrderDetail.as_view(), name='order-detail'),
    url(r'^orders/(?P<pk>[0-9]+)/lines/$', views.OrderLineList.as_view(), name='order-lines-list'),
    url(r'^orderlines/(?P<pk>[0-9]+)/$', views.OrderLineDetail.as_view(), name='order-lines-detail'),
    url(r'^orderlineattributes/(?P<pk>[0-9]+)/$', views.OrderLineAttributeDetail.as_view(), name='order-lineattributes-detail'),
    url(r'^countries/$', views.CountryList.as_view(), name='country-list'),
    url(r'^countries/(?P<pk>[A-z]+)/$', views.CountryDetail.as_view(), name='country-detail'),
    url(r'^partners/$', views.PartnerList.as_view(), name='partner-list'),
    url(r'^partners/(?P<pk>[0-9]+)/$', views.PartnerDetail.as_view(), name='partner-detail'),
    url(r'^useraddresses/$', views.UserAddressList.as_view(), name='useraddress-list'),
    url(r'^useraddresses/(?P<pk>[0-9]+)/$', views.UserAddressDetail.as_view(), name='useraddress-detail')
]

urlpatterns = format_suffix_patterns(urlpatterns)



import logging
from rest_framework import serializers

from oscarapi.basket import operations

from oscarapi.utils import (
    overridable,
    OscarModelSerializer,
    OscarHyperlinkedModelSerializer,
    DrillDownHyperlinkedIdentityField
)
from oscarapi.serializers.fields import TaxIncludedDecimalField

from django.utils.translation import ugettext as _
from oscar.core.loading import get_model
from decimal import Decimal

logger = logging.getLogger(__name__)

Basket = get_model('basket', 'Basket')
Line = get_model('basket', 'Line')
LineAttribute = get_model('basket', 'LineAttribute')
StockRecord = get_model('partner', 'StockRecord')
Voucher = get_model('voucher', 'Voucher')


class VoucherSerializer(OscarModelSerializer):
    class Meta:
        model = Voucher
        fields = overridable('OSCARAPI_VOUCHER_FIELDS', default=(
            'name', 'code', 'start_datetime', 'end_datetime'
        ))


class OfferDiscountSerializer(serializers.Serializer):
    description = serializers.CharField()
    name = serializers.CharField()
    amount = serializers.DecimalField(
        decimal_places=2, max_digits=12, source='discount')


class VoucherDiscountSerializer(OfferDiscountSerializer):
    voucher = VoucherSerializer(required=False)


class BasketSerializer(serializers.HyperlinkedModelSerializer):
    lines = serializers.HyperlinkedIdentityField(view_name='basket-lines-list')
    offer_discounts = OfferDiscountSerializer(many=True, required=False)
    total_excl_tax = serializers.DecimalField(
        decimal_places=2, max_digits=12, required=False)
    total_excl_tax_excl_discounts = serializers.DecimalField(
        decimal_places=2, max_digits=12, required=False)
    total_incl_tax = TaxIncludedDecimalField(
        excl_tax_field='total_excl_tax', decimal_places=2,
        max_digits=12, required=False)
    total_incl_tax_excl_discounts = TaxIncludedDecimalField(
        excl_tax_field='total_excl_tax_excl_discounts', decimal_places=2,
        max_digits=12, required=False)
    total_tax = TaxIncludedDecimalField(
        excl_tax_value=Decimal('0.00'), decimal_places=2,
        max_digits=12, required=False)
    currency = serializers.CharField(required=False)
    voucher_discounts = VoucherDiscountSerializer(many=True, required=False)

    class Meta:
        model = Basket
        fields = overridable('OSCARAPI_BASKET_FIELDS', default=(
            'id', 'owner', 'status', 'lines',
            'url', 'total_excl_tax',
            'total_excl_tax_excl_discounts', 'total_incl_tax',
            'total_incl_tax_excl_discounts', 'total_tax', 'currency',
            'voucher_discounts', 'offer_discounts', 'is_tax_known'))

    def get_validation_exclusions(self, instance=None):
        """
        This is needed because oscar declared the owner field as ``null=True``,
        but ``blank=False``. That means the validator will claim you can not
        leave this value set to None.
        """
        return super(BasketSerializer, self).get_validation_exclusions(
            instance) + ['owner']


class LineAttributeSerializer(OscarHyperlinkedModelSerializer):
    class Meta:
        model = LineAttribute
        fields = '__all__'


class BasketLineSerializer(OscarHyperlinkedModelSerializer):
    """
    This serializer computes the prices of this line by using the basket
    strategy.
    """
    url = DrillDownHyperlinkedIdentityField(
        view_name='basket-line-detail',
        extra_url_kwargs={'basket_pk': 'basket.id'})
    attributes = LineAttributeSerializer(
        many=True, fields=('url', 'option', 'value'),
        required=False, read_only=True)
    price_excl_tax = serializers.DecimalField(
        decimal_places=2, max_digits=12,
        source='line_price_excl_tax_incl_discounts')
    price_incl_tax = TaxIncludedDecimalField(
        decimal_places=2, max_digits=12,
        excl_tax_field='line_price_excl_tax_incl_discounts',
        source='line_price_incl_tax_incl_discounts')
    price_incl_tax_excl_discounts = TaxIncludedDecimalField(
        decimal_places=2, max_digits=12,
        excl_tax_field='line_price_excl_tax',
        source='line_price_incl_tax')
    price_excl_tax_excl_discounts = serializers.DecimalField(
        decimal_places=2, max_digits=12,
        source='line_price_excl_tax')
    warning = serializers.CharField(
        read_only=True, required=False, source='get_warning')

    @property
    def basket_pk(self):
        return self.kwargs.get('basket_pk')

    class Meta:
        model = Line
        fields = overridable('OSCARAPI_BASKETLINE_FIELDS', default=(
            'url', 'product', 'quantity', 'attributes', 'price_currency',
            'price_excl_tax', 'price_incl_tax',
            'price_incl_tax_excl_discounts', 'price_excl_tax_excl_discounts',
            'is_tax_known', 'warning', 'basket', 'stockrecord', 'date_created'
        ))

    def to_representation(self, obj):
        # This override is needed to reflect offer discounts or strategy
        # related prices immediately in the response
        operations.assign_basket_strategy(obj.basket, self.context['request'])

        # Oscar stores the calculated discount in line._discount_incl_tax or
        # line._discount_excl_tax when offers are applied. So by just
        # retrieving the line from the db you will loose this values, that's
        # why we need to get the line from the in-memory resultset here
        lines = (x for x in obj.basket.all_lines() if x.id == obj.id)
        line = next(lines, None)

        return super(BasketLineSerializer, self).to_representation(line)


class LineSerializer(serializers.HyperlinkedModelSerializer):
    """
    This serializer just shows fields stored in the database for this line.
    """
    attributes = LineAttributeSerializer(
        many=True,
        fields=('url', 'option', 'value'),
        required=False,
        read_only=True)

    class Meta:
        model = Line
        fields = '__all__'


class StockRecordSerializer(serializers.ModelSerializer):

    class Meta:
        model = StockRecord
        fields = '__all__'


class VoucherAddSerializer(serializers.Serializer):
    vouchercode = serializers.CharField(max_length=128, required=True)

    def validate(self, attrs):
        # oscar expects this always to be uppercase.
        attrs['vouchercode'] = attrs['vouchercode'].upper()

        request = self.context.get('request')
        try:
            voucher = Voucher.objects.get(code=attrs.get('vouchercode'))

            # check expiry date
            if not voucher.is_active():
                message = _("The '%(code)s' voucher has expired") % {
                    'code': voucher.code
                }
                raise serializers.ValidationError(message)

            # check voucher rules
            is_available, message = voucher.is_available_to_user(request.user)
            if not is_available:
                raise serializers.ValidationError(message)
        except Voucher.DoesNotExist:
            raise serializers.ValidationError(_('Voucher code unknown'))

        # set instance to the voucher so we can use this in the view
        self.instance = voucher
        return attrs

    def create(self, validated_data):
        return Voucher.objects.create(**validated_data)
